from urlparse import urlsplit, urlunsplit, urljoin, uses_fragment, uses_relative, uses_netloc
from cgi import parse_qsl
from urllib import quote_plus, urlencode
from copy import copy
from mext.dispatch import *
from mext.const import *

__all__ = ['Uri', 'RelativeUri', 'AbsoluteUri', 'urlquote', 'encode_query']

OTHER = const()

def encode_query(*args, **params):
    query = list(args)
    for key, value in params.iteritems():
        value_str = quote_plus(unicode(value).encode('utf-8'))
        query.append("%s=%s" % (key, value_str))
    return '&'.join(query)

def urlquote(*args, **params):
    path = u'/'.join(args).encode('utf8')
    path = quote_plus(path, '/')
    if not path.startswith('/'):
        path = '/' + path
    query = encode_query(**params)
    return (path + '?' + query) if query else path

scheme_requires_path = set(['http', 'https'])

class Uri(object):
    adapt = Adapter('__uri__')

    def __init__(self, *args):
        if len(args) == 1:
            args = urlsplit(unicode(args[0]))
        self.scheme, self.netloc, self.path, self.query, self.fragment = args
        if self.scheme and self.scheme not in uses_fragment:
            raise ValueError('URI scheme %r does not support fragment identifiers' % self.scheme)
        cls = AbsoluteUri if self._is_absolute else RelativeUri
        if not isinstance(self, cls):
            if not issubclass(cls, self.__class__):
                raise TypeError("%r is not a subclass of %r" % (cls, self.__class__))
            self.__class__ = cls
        if not self.path and self.scheme in scheme_requires_path:
            self.path = '/'

    def __str__(self):
        return urlunsplit((self.scheme, self.netloc, self.path, self.query, self.fragment))

    def __repr__(self):
        return self.__class__.__name__ + '(%s)' % repr(str(self))
        #return self.__class__.__name__ + '(%s)' % str.__repr__(self)

    @property
    def path_with_query(self):
        path = self.path
        if self.query:
            path += '?' + self.query
        return path

    @classmethod
    def from_parts(cls, scheme=None, netloc=None, path='/', query=None, fragment=None):
        return cls(scheme, netloc, path, query, fragment)

    def copy(self):
        return copy(self)

    def join(self, ref):
        return Uri(urljoin(str(self), str(ref)))

    #@@ maybe just use .join('#frag') ?
    def with_fragment(self, fragment):
        r = self.copy()
        r.fragment = fragment
        return r
        #return Uri(urlunsplit((self.scheme, self.netloc, self.path, self.query, frag)))

    def as_dir(self, save_query=False):
        r = self.copy()
        if not r.path.endswith('/'):
            r.path += '/'
        if not save_query:
            r.query = None
            r.fragment = None
        return r

    @property
    def _is_absolute(self):
        return bool(self.scheme #and self.path
            and (self.netloc or (self.scheme not in uses_netloc))
        )

    def defrag(self):
        if self.fragment:
            return self.with_fragment(None)
        else:
            return self

    def relative(self, base):
        base = Uri.adapt(base)
        if not base.is_absolute:
            raise ValueError("Cannon generate relative links from a relative base")
        elif (base.scheme or self.scheme) not in uses_relative:
            return self
        elif self.scheme == base.scheme and self.netloc == base.netloc:
            parts = ['', '', self.path, self.query, self.fragment]
            if base.path == self.path and base.query == self.query:
                parts[2] = parts[3] = ''
            elif '/' in self.path:
                base_split = base.path.rsplit('/', 1)
                self_split = self.path.rsplit('/', 1)
                if base_split[0] == self_split[0]:
                    parts[2] = self_split[1]
            ref = RelativeUri.from_parts(*parts)
            #assert base.join(ref) == self, repr((base, ref, base.join(ref), self))
            return ref
        else:
            return self


    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == other

    def __ne__(self, other):
        return not (self == other)


    def _get_query(self):
        if self._query is OTHER:
            self._query = urlencode(self._query_dict)
            self._query_dict = OTHER
        return self._query
    def _set_query(self, query):
        self._query = query
        self._query_dict = OTHER
    query = property(_get_query, _set_query)


    def _get_query_dict(self):
        if self._query_dict is OTHER:
            self._query_dict = dict(parse_qsl(self._query))
            self._query = OTHER
        return self._query_dict
    def _set_query_dict(self, query_dict):
        self._query = OTHER
        self._query_dict = query_dict
    query_dict = property(_get_query_dict, _set_query_dict)




Uri.adapt[basestring] = Uri
Uri.adapt.register('urllib2.Request', lambda req: AbsoluteUri(req.get_full_url()))

class AbsoluteUri(Uri):
    is_absolute = True

class RelativeUri(Uri):
    is_absolute = False

    def __str__(self):
        if (self.scheme or self.netloc or self.path or self.query):
            return Uri.__str__(self)
        else:
            return '#%s' % (self.fragment or '')

    def __nonzero__(self):
        return bool(self.scheme or self.netloc or self.path or self.query or self.fragment)


